import pika
import pickle
import traceback
import sys
from threading import Event

from isc import log

from gevent import sleep
from gevent import monkey, spawn
monkey.patch_all()


class Node(object):
    def __init__(self):
        self.services = {}
        self.listeners = {}
        self._is_ready = Event()
        self.params = pika.ConnectionParameters('localhost')

    def on_message(self, channel, method, properties, body):
        spawn(self.validate_message, channel, method, properties, body)

    def validate_message(self, channel, method, properties, body):
        service_name = method.routing_key[12:]

        if service_name not in self.services:  # pragma: no cover
            return

        channel.basic_ack(delivery_tag=method.delivery_tag)

        service = self.services[service_name]
        result = self.call_service(service, body)

        channel.basic_publish(exchange='', routing_key=properties.reply_to, properties=pika.BasicProperties(
            correlation_id=properties.correlation_id
        ), body=pickle.dumps(result))

    def call_service(self, service, body):
        try:
            fn_name, args, kwargs = pickle.loads(body)
            fn = getattr(service, fn_name, None)

            if fn is None:
                log.warning('No method %s', fn_name)
                raise Exception('Could not find method {}.'.format(fn_name))

            if getattr(fn, '__exposed__', None) is None:
                log.warning('Method %s is not exposed', fn_name)
                raise Exception('You are not allowed to call unexposed method {}.'.format(fn_name))
        except Exception as e:
            return (str(e), None)
        else:
            try:
                result = (None, fn(*args, **kwargs))
                log.debug('{}(*{}, **{})'.format(fn_name, args, kwargs))
                return result
            except Exception as e:
                tb = sys.exc_info()[2]
                frame = traceback.extract_tb(tb)[-1]
                if isinstance(frame, tuple):  # pragma: no cover
                    filename, lineno, _, line = frame
                else:  # pragma: no cover
                    filename, lineno, line = frame.filename, frame.lineno, frame.line
                log.error('Error in RPC method "{}", file {}:{}:\n    {}\n{}: {}'.format(fn_name, filename, lineno, line, e.__class__.__name__, str(e)))
                return (str(e), None)

    def on_broadcast(self, channel, method, properties, body):
        # TODO: Handle decoding errors
        event, data = pickle.loads(body)

        listeners = self.listeners.get(event, [])
        for fn in listeners:
            spawn(fn, data)

    def register_service(self, service):
        if service.name in self.services:
            raise Exception('Service {} is already registered.'.format(service.name))
        self.services[service.name] = service

    def run(self):
        while True:
            try:
                conn = pika.BlockingConnection(self.params)
                self.conn = conn
                #     conn = pika.BlockingConnection(pika.ConnectionParameters('localhost'))
                # except Exception as e:
                #     print(str(e))
                #     print('RabbitMQ not running? Retrying connection in 1 second.')
                #     sleep(1)
                #     continue

                channel = conn.channel()
                for service in self.services.values():
                    queue = 'isc_service_{}'.format(service.name)
                    channel.queue_declare(queue=queue)
                    channel.basic_consume(self.on_message, queue=queue, no_ack=False)

                    for attr in [getattr(service, attr, None) for attr in dir(service)]:
                        events = getattr(attr, '__on__', None)
                        if events:
                            for event in events:
                                if event not in self.listeners:
                                    self.listeners[event] = []
                                self.listeners[event].append(attr)

                channel.exchange_declare(exchange='isc_fanout', type='fanout')
                fanout_queue = channel.queue_declare(exclusive=True)
                channel.queue_bind(exchange='isc_fanout', queue=fanout_queue.method.queue)

                channel.basic_consume(self.on_broadcast, queue=fanout_queue.method.queue, no_ack=True)
                log.info('Ready')
                self.channel = channel
                self._is_ready.set()
                channel.start_consuming()
                break
            except pika.exceptions.ConnectionClosed:  # pragma: no cover
                self._is_ready.clear()
                log.error('Connection closed, retrying in 3 seconds')
                sleep(3)
                continue

    def wait_for_ready(self):
        self._is_ready.wait()

    def stop(self):
        self.conn.add_timeout(0, lambda: self.channel.stop_consuming())


def expose(fn):
    fn.__exposed__ = True
    return fn


def on(event):
    def wrapper(fn):
        if getattr(fn, '__on__', None) is None:
            fn.__on__ = set()
        fn.__on__ |= set([event])
        return fn
    return wrapper



from gevent import monkey, spawn
monkey.patch_all()

import pika
import pickle
import uuid
from threading import Event

from isc import log


class IPCException(Exception):
    pass


class RemoteException(IPCException):
    pass


class LocalException(IPCException):
    pass


class TimeoutException(LocalException):
    pass


class FutureResult(object):
    """
    Encapsulates future result. Provides interface to wait for future data.
    """
    def __init__(self):
        self.event = Event()
        self.result = None

    def wait(self, timeout=5):
        """
        Blocks until data is ready.
        """
        self.event.wait(timeout=float(timeout))
        return self.result

    def happen(self, result):
        """
        Fulfills the result and sets "ready" event.
        """
        self.result = result
        self.event.set()


class Connection(object):
    """
    Represents a single low-level connection to the ISC messaging broker.
    """
    def __init__(self, host):
        self.host = host
        self.future_results = {}

    def connect(self):
        """
        Connect to broker and create a callback queue with "exclusive" flag.
        """
        try:
            self.conn = pika.BlockingConnection(pika.ConnectionParameters(self.host))
        except:
            log.error('RabbitMQ not running?')
            return False

        self.channel = self.conn.channel()
        # self.channel.queue_declare(queue='test')

        self.callback_queue = self.channel.queue_declare(exclusive=True).method.queue

        self.channel.basic_consume(self.on_response, no_ack=True, queue=self.callback_queue)

        return True

    def start_consuming(self):
        """
        Start consuming messages.
        This function is blocking.
        """
        self.channel.start_consuming()

    def stop_consuming(self):
        self.channel.stop_consuming()

    def on_response(self, channel, method, properties, body):
        """
        Called when a message is consumed.
        """
        future_result = self.future_results.get(properties.correlation_id, None)
        if not future_result:  # pragma: no cover
            # TODO: Should not happen!
            log.error('FIXME: This should not happen.')
            return
        try:
            exception, result = pickle.loads(body)
            if exception:
                exception = RemoteException(exception)
        except Exception as e:  # pragma: no cover
            exception, result = LocalException(str(e)), None
        future_result.happen((exception, result))

    def call(self, service, method, *args, **kwargs):
        """
        Serialize & publish method call request.
        """
        corr_id = str(uuid.uuid4())

        future_result = FutureResult()
        self.future_results[corr_id] = future_result

        self.channel.basic_publish(
            exchange='', routing_key='isc_service_{}'.format(service), body=pickle.dumps((method, args, kwargs)),
            properties=pika.BasicProperties(
                reply_to=self.callback_queue, correlation_id=corr_id
            )
        )

        return future_result

    def notify(self, event, data):
        """
        Serialize & publish notification.
        """
        corr_id = str(uuid.uuid4())

        self.channel.basic_publish(
            exchange='isc_fanout', routing_key='', body=pickle.dumps((event, data)),
            properties=pika.BasicProperties(
                correlation_id=corr_id
            )
        )


class ServiceProxy(object):
    def __init__(self, client, service_name):
        self.client = client
        self.service_name = service_name

    def __getattr__(self, attr):
        return MethodProxy(self.client, self.service_name, attr)


class MethodProxy(object):
    def __init__(self, client, service_name, method_name):
        self.client = client
        self.service_name = service_name
        self.method_name = method_name

    def __call__(self, *args, **kwargs):
        return self.client.invoke(self.service_name, self.method_name, *args, **kwargs)


class Client(object):
    """
    Provides simple interface to the Connection.
    """
    def __init__(self, hostname='127.0.0.1', timeout=5):
        self.connection = Connection(hostname)
        self.timeout = timeout

    def connect(self):
        if not self.connection.connect():
            return False

        spawn(self.connection.start_consuming)

    def stop(self):
        self.connection.stop_consuming()

    def set_timeout(self, timeout):
        self.timeout = timeout

    def invoke(self, service, method, *args, **kwargs):
        """
        Call a remote method.
        """

        future_result = self.connection.call(service, method, *args, **kwargs)

        data = future_result.wait(self.timeout)

        if data:
            exception, result = data
        else:
            exception, result = TimeoutException('Method {}.{} timed out.'.format(service, method)), None

        if exception:
            raise exception
        else:
            return result

    def notify(self, event, data):
        """
        Publish a notification.
        """
        self.connection.notify(event, data)

    def __getattr__(self, attr):
        return ServiceProxy(self, attr)



from unittest import TestCase
from time import sleep
from gevent import spawn
from gevent.event import Event
from .server import Node, expose, on
from .client import Client, RemoteException, TimeoutException


class ExampleService(object):
    name = 'example'

    def __init__(self):
        # Just for the tests, don't do this - services *must* be stateless
        self.stuff_done_event = Event()

    @expose
    def add(self, a, b):
        return a + b

    @expose
    def raise_error(self):
        raise Exception('testing')

    def private_method(self):  # pragma: no cover
        return 'Cannot call me!'

    @on('boom')
    def do_stuff(self, arg):
        print('Got stuff:', arg)
        self.stuff_done_event.set()

    @expose
    def slow_method(self):  # pragma: no cover
        sleep(3)


class GenericTest(TestCase):
    def setUp(self):
        self.node = Node()
        self.service = ExampleService()
        self.node.register_service(self.service)
        self.node_greenlet = spawn(self.node.run)
        self.node.wait_for_ready()
        self.client = Client()
        self.client.connect()

    def tearDown(self):
        self.client.stop()
        self.node.stop()
        self.node_greenlet.join()

    def test_method_success(self):
        self.assertEqual(self.client.example.add(2, 3), 5)
        self.assertEqual(self.client.invoke('example', 'add', 2, 3), 5)

    def test_method_exceptions(self):
        self.assertRaises(RemoteException, self.client.example.add)
        self.assertRaises(RemoteException, self.client.example.add, (2, '3'))

    def test_unexisting_method(self):
        self.assertRaises(RemoteException, self.client.example.unexisting_method)
        # self.assertRaises(RemoteException, self.client.unexisting_service.add, (2, 3))

    def test_unexisting_service(self):
        self.client.set_timeout(1)
        self.assertRaises(TimeoutException, self.client.unexisting_service.some_method)

    def test_raises_exception(self):
        self.assertRaises(RemoteException, self.client.example.raise_error)
        try:
            self.client.example.raise_error()
        except RemoteException as e:
            self.assertEqual(str(e), 'testing')

    def test_private_method(self):
        self.assertRaises(RemoteException, self.client.example.private_method)

    def test_notify(self):
        self.client.notify('boom', dict(place='some_place'))
        self.assertEqual(self.service.stuff_done_event.wait(3), True)

    def test_register_again(self):
        self.assertRaises(Exception, self.node.register_service, self.service)

    def test_connection_failed(self):
        client = Client('unexisting.hostname.it.should.not.exist.and.if.it.does.then.this.test.will.fail')
        self.assertFalse(client.connect())

    # def test_slow_method(self):
    #     # TODO: Does not pass.
    #     self.client.set_timeout(1)
    #     self.assertRaises(TimeoutException, self.client.example.slow_method)
